package TSim;
import java.util.Random;

/** The TSimInterface is the intended interface between TSim and the 
 *  laboration. It handles the extraction of information from TSim and
 *  provides methods for manipulating trains, sensors and switches.
 *
 */
public class TSimInterface implements Runnable
{
    private final Random random=new Random();
    /** Switch state: switch points to the left. */
    public static final int SWITCH_LEFT  = 0x01;
    /** Switch state: switch points to the right. */
    public static final int SWITCH_RIGHT = 0x02;
    
    SynchronizedFIFO commandFIFO     = new SynchronizedFIFO(10);

    TrainEvent lastEvent=null;

    SynchronizedFIFO trainSensorFIFO=new SynchronizedFIFO(10);

    TSimStream sTSim;

    private int semaphores=0;

    /** Create a new TSimInterface
     *
     */
    TSimInterface(int port)
    {
	sTSim = new TSimStream(port);
    }
        
    public void run() 
    {
	while (true) {
	    
	    try {
		TSimInformation dInfo = sTSim.read();
		
		if (dInfo instanceof CommandStatus)
		    commandFIFO.put(dInfo);
		
		/* if we got a train event something went wrong. 
		 */

		else if (dInfo instanceof TrainEvent) {
		    TrainEvent tE = (TrainEvent) dInfo;
		    
		    /* Store the error event to make future commands 
		       result in an exception.
		    */
		    
		    lastEvent=tE;
		    
		    /* and report the error event on standard error
		     */
		    
		    reportTrainEvent(tE);

		}

		else if (dInfo instanceof SensorEvent) {
		    SensorEvent sEvent = (SensorEvent) dInfo;
		    
		    trainSensorFIFO.put(sEvent);
		}
	    }
	    catch (UnparsableInputException e) {
		System.err.println(e.getMessage());
		System.exit(1);
	    }
	}
    }

    /** Helper method that reports an error event on standard error.
     *
     *  @param e   the error event.
     *
     */
    void reportTrainEvent(TrainEvent e) 
    {
	System.err.println(e);
    }

    /** Set train target speed to the given value. The train's speed does not
     * change immediately and the method call does not wait until the train
     * has reached the target speed.
     *
     *  @param speed    the new speed of the train.
     *  @throws CommandException if the train has encountered a fatal error
     *
     */
    public synchronized void setSpeed(int speed)
	throws CommandException
    {
	TrainEvent tE = lastEvent;
	
	/* some event has happened for this train i.e. it has crashed */
	
	if (tE != null)
	    throw new CommandException(tE.toString());

	sTSim.write("SetSpeed " + String.valueOf(speed));
	
	CommandStatus cStat = (CommandStatus) commandFIFO.get();

	if (cStat.getStatus() != CommandStatus.OK)
	    throw new CommandException(cStat.toString());

    }

    /** Load and unload as necessary at station. This should be called at
     * stations. The train is stopped before loading/unloading starts.
     *
     *  @throws CommandException  if the train has encountered a fatal error
     *
     */
    public synchronized void loadUnload()
	throws CommandException
    {
	TrainEvent tE = lastEvent;
	
	/* some event has happened for this train i.e. it has crashed */
	
	if (tE != null)
	    throw new CommandException(tE.toString());

	float wt;
	if (random.nextInt(10)==0)
	    wt=random.nextFloat()*(float)140.0;
	else
	    wt=random.nextFloat()*(float)20.0;
	sTSim.write("LoadUnload 4 " + wt);
	
	CommandStatus cStat = (CommandStatus) commandFIFO.get();

	if (cStat.getStatus() != CommandStatus.OK)
	    throw new CommandException(cStat.toString());
    }

    /** Set the position of the specified switch. Valid positions are
     *  SWITCH_LEFT and SWITCH_RIGHT.
     * 
     *  @param xPos   the x coordinate of the switch.
     *  @param yPos   the y coordinate of the switch.
     *  @param switchPos  the new position of the switch.
     *
     *  @throws CommandException if the coordinates of the switch were invalid
     *                          (NO_SUCH_SWITCH) or if there was a train on
     *                          the switch (TRAIN_ON_SWITCH)
     *
     */
    public synchronized void setSwitch(int xPos, int yPos, int switchPos)
	throws CommandException
    {
	
	sTSim.write("SetSwitch " + String.valueOf(xPos) + " " +
		    String.valueOf(yPos) + 
		    (switchPos == SWITCH_LEFT ?
		     " LeftSwitch" : " RightSwitch"));
	
	CommandStatus cStat = (CommandStatus) commandFIFO.get();
	
	if (cStat.getStatus() != CommandStatus.OK)
	    throw new CommandException(cStat.toString());
    }

    
    /** Wait until the train passes a sensor.
     *
     *  @return a SensorEvent representing the information about the event
     *  @throws  CommandException if the train has encountered a fatal error.
     *
     */
    public SensorEvent getSensor() throws CommandException
    {
	TrainEvent tE = lastEvent;
	
	/* some event has happened for this train i.e. it has crashed */
	
	if (tE != null)
	    throw new CommandException(tE.toString());

	return (SensorEvent) trainSensorFIFO.get();
    }

    /**
     * Create a semaphore with specified starting value. The semaphore is shared
     * with all other trains on the track; the nth Semaphore object created by
     * each train corresponds to the same semaphore.
     * @param s Initial value to give semaphore (if not already created).
     * @return The new semaphore.
     */
    public Semaphore getSemaphore(int s) {
	return new Semaphore(semaphores++, s, this);
    }

}
