package TSim;

import java.io.*;
import java.net.*;

/** Reads text from the given input stream connected to TSim and
 *  returns a stream of TSimInformation.
 *
 **/

class TSimStream 
{
    static boolean logging;
    StreamTokenizer sTokenizer;
    PrintStream out;
    BufferedReader inReader;
    Socket sock;

    /** Creates a new TSimStream that reads from the given InputStream.
     * 
     *  @param in the input stream to read text from.
     *
     */

    public TSimStream(int port)
    {
        try {
	    sock=new Socket();
	    sock.setTcpNoDelay(true);
	    // Removed for Java 1.4 compatibility; doesn't seem to help much.
//	    sock.setPerformancePreferences(0,1,0);
	    sock.connect(new InetSocketAddress(
			     InetAddress.getByName("127.0.0.1"),
			     port));
            inReader=new BufferedReader(new InputStreamReader(
					    sock.getInputStream()));
            out=new PrintStream(sock.getOutputStream());
        } catch(Exception e) {
            throw new RuntimeException("Failed to connect", e);
        }

	sTokenizer = new StreamTokenizer(inReader);
    }

    /** Writes to output stream.
     * @ param s String to write
     */
    public void write(String s) {
	if (logging) {
	    System.err.println("sending command "+s);
	    Thread.currentThread().dumpStack();
	}
	out.println(s);
	out.flush();
    }
   
    /** Reads from the input stream and returns a class representing
     *  the message from TSim.
     *
     *  @return 
     *  @throws UnparsableInputException when the input from TSim is 
     *                                   not parsable.
     **/
 
    public TSimInformation read() throws UnparsableInputException
    {
	String cmd = readString(sTokenizer);
	
	TSimInformation info = null;

	if (cmd.equals("SuccessStatus"))
	    info = new CommandStatus(CommandStatus.OK);
	else if (cmd.equals("NoSwitchStatus"))
	    info = new CommandStatus(CommandStatus.NO_SUCH_SWITCH);
	else if (cmd.equals("IllegalTrainNoStatus"))
	    info = new CommandStatus(CommandStatus.NO_SUCH_TRAIN);
	else if (cmd.equals("IllegalReverseStatus"))
	    info = new CommandStatus(CommandStatus.ILLEGAL_SPEED);
	else if (cmd.equals("IllegalRailPosStatus"))
	    info = new CommandStatus(CommandStatus.NO_SUCH_POS);
	else if (cmd.equals("IllegalSwitchingStatus"))
	    info = new CommandStatus(CommandStatus.TRAIN_ON_SWITCH);
	else if (cmd.equals("Sensor")) {
	    
	    int xPos = readInt(sTokenizer);
	    int yPos = readInt(sTokenizer);
	    
	    int sStatus = SensorEvent.ACTIVE;
	    
	    if (readString(sTokenizer).equals("inactive"))
		sStatus = SensorEvent.INACTIVE;
	    
	    info = new SensorEvent(xPos, yPos, sStatus);
	    
	}
	else if (cmd.equals("TrainCollisionEvent")) {
	    
	    info = new TrainEvent(TrainEvent.TRAIN_COLLISION);
	}
	else if (cmd.equals("StopCollisionEvent")) {
	    info = new TrainEvent(TrainEvent.STOP_COLLISION);
	}
	else if (cmd.equals("DerailmentEvent")) {
	    info = new TrainEvent(TrainEvent.DERAILMENT);
	}
	else if (cmd.equals("SemaphoreNotFree"))
	    info = new CommandStatus(CommandStatus.SEMAPHORE_NOT_FREE);
	else if (cmd.equals("IllegalSemaphore"))
	    info = new CommandStatus(CommandStatus.NO_SUCH_SEMAPHORE);
	else if (cmd.equals("TrainNotAtStation"))
	    info = new CommandStatus(CommandStatus.TRAIN_NOT_AT_STATION);
	else if (cmd.equals("TimeoutEvent"))
	    info = new CommandStatus(CommandStatus.TIMEOUT_EVENT);
	else throw new UnparsableInputException
	    ("unknown information from tsim "+cmd);
	
	/* What if there is something that would result in a TT_EOL.
	   Shouldn't we handle this here?
	*/

	return info;
	
    }
    
    /** Help method that reads a string from the tokenizer and
     *  returns it or throws an exception if the next token was not 
     *  a string.
     *
     *  @param sTokenizer the stream tokenizer to read from.
     *  @returns the read string.
     *  @throws UnparsableInputException when the next token was not 
     *                                   a string.
     */


    String readString(StreamTokenizer sTokenizer) throws
    UnparsableInputException
    {
	try {
	    int type = sTokenizer.nextToken();
	    
	    if (type == StreamTokenizer.TT_EOF)
		throw new UnparsableInputException
		    ("simulator connection closed");

	    if (type != StreamTokenizer.TT_WORD)
		throw new UnparsableInputException
		    ("string expected");
	    
	    return sTokenizer.sval;
	}
	catch (IOException e) {
	    throw new UnparsableInputException(e.getMessage());
	}
	
    }

    /** Help method that reads an integer from the tokenizer and
     *  returns it or throws an exception if the next token was not 
     *  an integer.
     *
     *  @param sTokenizer the stream tokenizer to read from.
     *  @returns the read integer.
     *  @throws UnparsableInputException when the next token was not 
     *                                   an integer.
     */

    int readInt(StreamTokenizer sTokenizer) throws
    UnparsableInputException
    {
	try {
	    
	    int type = sTokenizer.nextToken();

	    if (type != StreamTokenizer.TT_NUMBER)
		throw new UnparsableInputException
		    ("int expected");
	    
	    return (int) sTokenizer.nval;
	}
	catch (IOException e) {
	    throw new UnparsableInputException(e.getMessage());
	}
    }
    
}








