#include <stdio.h>
#include <strings.h>
#if defined sun3 || defined sun4
#include <sys/filio.h>
#endif
#ifdef SOLARIS
#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#else
#include <sys/ioctl.h>
#endif

#include <string.h>
#include <stdlib.h>

#include <unistd.h>                     /* for read on LINUX */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <X11/Xlib.h>			/* for Xlib definitions */
#include <X11/StringDefs.h>		/* for useful atom names */
#include <X11/Intrinsic.h>

#include "train.h"
#include "trainP.h"
#include "file.h"
#include "semaphores.h"
#include "resources.h"
#include "netinet/tcp.h"

#define BUFSIZE 512

/* prototypes for gcc */
extern int errcheck(int i, int ec, char *s);
void Terminate (int i);

#ifndef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#endif

#define SEPS " \r\t\n"
  static String seps = SEPS;

void Acknowledge(tTrain *train, tStatus status)
{
    if (status!=NoStatus) {
	fprintf(train->out_file, "%s\n", StatusToString[status]);
	if (use_log()) {
	    fprintf(stderr, "%d: %s\n", train->no,
		    StatusToString[status]);
	}
	fflush(train->out_file);
    }
}

static tStatus ParseFloat(float *f)
{
  String value;
  if(!(sscanf(value = strtok(NULL, seps), "%f", f)))
    fprintf(stderr, "Warning: Illegal float value: \"%s\"\n", value);
  return(SuccessStatus);
}

static tStatus ParseInt(int *i)
{
  String value = strtok(NULL, seps);
  if(!(value && sscanf(value, "%d", i)))
    fprintf(stderr, "Warning: Illegal integer value: \"%s\"\n", value);
  return(SuccessStatus);
}

static tStatus ParsePos(tRailPos *pos)
{
  tStatus status;
  if((status = ParseInt(&pos->x))) return(status);
  if((status = ParseInt(&pos->y))) return(status);
  return(SuccessStatus);
}

static tStatus ParseSwitchState(tSwitchState *state)
{
  *state = StringToSwitchState(strtok(NULL, seps));
  return(SuccessStatus);
}

static tStatus SetSwitchCmd(tTrain *train, String command_line)
{
  tTrainLine train_line=train->line;
  tRailPos pos;
  tSwitchState state;
  tStatus status;

  if((status = ParsePos(&pos))) return(status);
  if((status = ParseSwitchState(&state))) return(status);
  return(SetSwitch(train_line, pos, state));
}

static tStatus SetSpeedCmd(tTrain *train, String command_line)
{
  tTrainNo train_no=train->no;
  tSpeed speed;
  tStatus status;

  if((status = ParseFloat(&speed))) return(status);
  return(SetSpeed(train->line, train_no, speed));
}

static tStatus LoadUnloadCmd(tTrain *train, String command_line)
{
  tTrainNo train_no=train->no;
  tSpeed tim;
  tStatus status;
  int dist;

  if((status = ParseInt(&dist))) return(status);
  if((status = ParseFloat(&tim))) return(status);
  if((status = SetSpeed(train->line, train_no, 0.0))) return(status);
  train->loading = tim/timeout_interval();
  if (train->loading<=0) train->loading=1;
  train->load_distance = dist;
  return NoStatus;
}

static tStatus CreateSemaphoreCmd(tTrain *train, String command_line) {
    tStatus status;
    int val,sem;
    if((status = ParseInt(&val))) return(status);
    if((status = ParseInt(&sem))) return(status);
    return(CreateSemaphore(train->line, sem, val));
}

static tStatus AcquireSemaphoreCmd(tTrain *train, String command_line) {
    tStatus status;
    int sem;
    if((status = ParseInt(&sem))) return(status);
    return(AcquireSemaphore(train, sem));
}

static tStatus ReleaseSemaphoreCmd(tTrain *train, String command_line) {
    tStatus status;
    int sem;
    if((status = ParseInt(&sem))) return(status);
    return(ReleaseSemaphore(train->line, sem));
}

static tStatus TrySemaphoreCmd(tTrain *train, String command_line) {
    tStatus status;
    int sem;
    if((status = ParseInt(&sem))) return(status);
    return(TrySemaphore(train->line, sem));
}

static void Dispatch(train, command_line)
     tTrain *train;
     String command_line;
{
  typedef struct {
    String name;
    tStatus (*procedure)();
  } tCommandPair;
  static tCommandPair commands[] = {
    { "SetSwitch", SetSwitchCmd },
    { "SetSpeed", SetSpeedCmd },
    { "CreateSemaphore", CreateSemaphoreCmd },
    { "AcquireSemaphore", AcquireSemaphoreCmd },
    { "ReleaseSemaphore", ReleaseSemaphoreCmd },
    { "LoadUnload", LoadUnloadCmd },
    { "TrySemaphore", TrySemaphoreCmd } };
  static int no_of_commands = 7;
  String command;
  int cmdno;

  if (use_log())
      fprintf(stderr, "%d: %s", train->no, command_line);

  command = strtok(command_line, seps);  /* Remove the command */
  if(!command) return;
  for(cmdno = 0; cmdno < no_of_commands; cmdno++)
    if(!strcasecmp(commands[cmdno].name, command)) {
      Acknowledge(train,
		  commands[cmdno].procedure(train, command_line));
      return;
    }
  Acknowledge(train, SuccessStatus);
  fprintf(stderr, "Warning: Illegal command received: %s\n", command);
}

static void ReadHandler(train, infiledesc, input_id)
     tTrain *train;
     int *infiledesc;
     XtInputId *input_id;
{
  int no_of_bytes;
  char *s;
  char *buf=train->buffer;

#ifdef SOLARIS
  errcheck(ioctl(*infiledesc, I_NREAD, &no_of_bytes), -1, "ioctl error");
#else
  errcheck(ioctl(*infiledesc, FIONREAD, &no_of_bytes), -1, "ioctl error");
#endif
  if(!no_of_bytes) /* Nothing to read - done! */
      if (train->in_file!=-1) {
	  XtRemoveInput(train->input_id);
	  fclose(train->out_file);
	  train->in_file=-1;
	  free(train->buffer);
	  if (strict())
	      Terminate(1);
	  return;
      }
  no_of_bytes = MIN(no_of_bytes, BUFSIZE - train->charsinbuf - 1);
  read(*infiledesc, buf+train->charsinbuf, no_of_bytes);
  buf[train->charsinbuf+no_of_bytes] = '\0';
  for(s=buf; ((*s)!='\0')&&(((*s)!='\n')); s++);
  if ((*s)=='\0') {
      /* String ends before newline; not a full command. */
      train->charsinbuf=s-buf;
      return;
  }

  /* A complete line has been received (and no more due to explicit acks). */
  Dispatch(train, buf);
  train->charsinbuf=0;
}

void SendSensorActive(train, pos, active)
     tTrain *train;
     tRailPos pos;
     Bool active;
{
  if(train->in_file!=-1) {
    fprintf(train->out_file, "Sensor %d %d ", pos.x, pos.y);
    if(active)
      fprintf(train->out_file, "active\n");
    else
      fprintf(train->out_file, "inactive\n");
    fflush(train->out_file);
  }
  if (use_log()) {
      fprintf(stderr, "%d: Sensor %d %d ", train->no, pos.x, pos.y);
      if(active)
	  fprintf(stderr, "active\n");
      else
	  fprintf(stderr, "inactive\n");
  }
}

void SendTrainEvent(train, event)
     tTrain *train;
     tEvent event;
{
    if((train->in_file!=-1) && event) {
	fprintf(train->out_file, "%s\n", EventToString[event]);
	fflush(train->out_file);
    }
    if (use_log()) {
	fprintf(stderr, "%d: %s\n", train->no, EventToString[event]);
    }
	
    if (strict())
	Terminate(0);
}

#define MAX_PENDING_CONNECTIONS 10

static void PipeHandler(train_line, infiledesc, input_id)
     tTrainLine train_line;
     int *infiledesc;
     XtInputId *input_id;
{
    tTrain* t=train_line->trains;
    XtAppContext apc=train_line->apc;
    int flag=1;
    int newsock=errcheck(accept(*infiledesc, NULL, NULL), 
			 -1, "connect error:");

    setsockopt(newsock, IPPROTO_TCP, TCP_NODELAY, 
	       &flag, sizeof(int));

    while((t!=NULL)&&(t->in_file!=-1))
	t=t->next;
    
    if (t==NULL) {
	fprintf(stderr, "tsim: Too many network clients; add more trains.");
	close(newsock);
	if (strict())
	    Terminate(1);
	return;
    }
    t->out_file = fdopen(newsock, "w");
    t->in_file = newsock;
    t->buffer=malloc(sizeof(char)*BUFSIZE);
    t->charsinbuf=0;
    setbuf(t->out_file, NULL);
    t->input_id=XtAppAddInput(apc, newsock, (XtPointer) XtInputReadMask, 
			      ReadHandler, t);
}

void AddListenSocket(apc, train_line)
     XtAppContext apc;
     tTrainLine train_line;
{
    struct sockaddr_in addr;
/*    int so_reuseaddress = 1;*/
    socklen_t addr_size=sizeof(struct sockaddr_in);
    bzero(&addr, addr_size);

    addr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
    addr.sin_family=AF_INET;
    addr.sin_port=htons(0);
    train_line->listen_socket=errcheck(socket(PF_INET, SOCK_STREAM, 0), -1,
				       "listen error (socket)");
/*    setsockopt(train_line->listen_socket, SOL_SOCKET, SO_REUSEADDR, 
      &so_reuseaddress, sizeof(so_reuseaddress));*/
    errcheck(bind(train_line->listen_socket, (struct  sockaddr *)(&addr),
		  sizeof(struct sockaddr_in)), -1,
	     "listen error (bind)");
    errcheck(listen(train_line->listen_socket, MAX_PENDING_CONNECTIONS), -1,
	     "listen error (listen)");
    train_line->apc=apc;

    train_line->input_id=XtAppAddInput(apc, train_line->listen_socket,
				       (XtPointer) XtInputReadMask, 
				       PipeHandler, train_line);
    errcheck(getsockname(train_line->listen_socket, (struct sockaddr *)(&addr),
			 &addr_size), -1, "getsockname error");
    fprintf(stdout, "%d\n", ntohs(addr.sin_port));
    fflush(stdout);
}

void StopNetworkTraffic (tTrainLine train_line) {
    tTrain* t=train_line->trains;
    XtRemoveInput(train_line->input_id);
    close(train_line->listen_socket);
    while(t!=NULL) {
	if (t->in_file!=-1) {
	    XtRemoveInput(t->input_id);
	    fclose(t->out_file);
	    free(t->buffer);
	}
	t=t->next;
    }
}
    
